/****************************************************************************

**  Gcssamer - Lightweight CSS Inspector for Firefox

**  Copyright (c) 2008  Charles Melhorn

**

**  This file is part of Gcssamer.

**

**  Gcssamer is free software; you can redistribute it and/or modify

**  it under the terms of the GNU General Public License as published by

**  the Free Software Foundation; either version 2 of the License, or

**  (at your option) any later version.

**

**  Gcssamer is distributed in the hope that it will be useful,

**  but WITHOUT ANY WARRANTY; without even the implied warranty of

**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

**  GNU General Public License for more details.

**

**  You should have received a copy of the GNU General Public License

**  along with this program; if not, write to the Free Software

**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**

**  Or see the GNU web site at: http://www.gnu.org

**

****************************************************************************/ 



// extend Array object
if (!Array.hasValue)
  Array.prototype.hasValue = function(aValue)
  {
    for(var idx=0; idx<this.length; idx++)
     {
       if (this[idx] == aValue)
         return true
     }
    return false;
  };
  

// globals, constants, services
var   gOriginalTargetElement = null;
var   gCurrentElement = null;
const gClipboardHelper = Components.classes["@mozilla.org/widget/clipboardhelper;1"]
                         .getService(Components.interfaces.nsIClipboardHelper);
const COL_SELECTEDFLAG = 0, COL_PROPSTRING = 1, COL_PROPVALUE = 2; // indices for style_data array



function truncateString(aString, aLength)
{
  return aString.substr(0, aLength) + (aString.length > aLength ? "..." : "");
}


function initAttributeInfo(aElement)
{
  var attribute_lbox = document.getElementById('attribute-lbox');

  // clear listbox
  for (var i=attribute_lbox.itemCount; i > 0; i--)
    attribute_lbox.removeItemAt(0);

  // populate listbox with attribute info
  for (var i=0; i < aElement.attributes.length; i++)
   {
    li = document.createElement('listitem');

    lc = document.createElement('listcell');
    lc.setAttribute('label', aElement.attributes.item(i).name);
    li.appendChild(lc);

    lc = document.createElement('listcell');
    lc.setAttribute('label', aElement.attributes.item(i).value);
    li.appendChild(lc);

    attribute_lbox.appendChild(li);
   }
}


function initCSSInfo(aElement)
{
  cssInfoTreeView.clear();

  document.getElementById('target-tag-type').value = aElement.tagName;
  document.getElementById('target-innerhtml').value = truncateString(aElement.innerHTML, 50);

  var style_string = "";
  var prop_string = "";

  // get ComputedStyle object for our element - in unfocused state
  aElement.blur();
  cs = aElement.ownerDocument.defaultView.getComputedStyle(aElement, null);

  // get user-preferred css properties
  var favorite_props = gGCSSAMERUserPrefUtils.getPref('cssprops');

  if (favorite_props == "")
    favorite_props = [];
  else
    favorite_props = favorite_props.split(';');

  var propset = document.getElementById('propset-rgroup').selectedItem.value;

  if (propset == "selected")
   {
     for (var propidx=0; propidx < favorite_props.length; propidx++)
      {
        prop_string = favorite_props[propidx];
        cssInfoTreeView.style_data[propidx]= new Array(3);
        cssInfoTreeView.style_data[propidx][COL_SELECTEDFLAG] = true;
        cssInfoTreeView.style_data[propidx][COL_PROPSTRING] = prop_string;
        cssInfoTreeView.style_data[propidx][COL_PROPVALUE] = cs.getPropertyValue(prop_string);
      }
   }
  else
   {
     for (var propidx=0; propidx < cs.length; propidx++)
      {
        prop_string = cs.item(propidx);
        cssInfoTreeView.style_data[propidx]= new Array(3);
        cssInfoTreeView.style_data[propidx][COL_SELECTEDFLAG] = favorite_props.hasValue(prop_string);
        cssInfoTreeView.style_data[propidx][COL_PROPSTRING] = prop_string;
        cssInfoTreeView.style_data[propidx][COL_PROPVALUE] = cs.getPropertyValue(prop_string);
      }
   }

  // force tree to redraw
  cssInfoTreeView.tree.rowCountChanged(0, cssInfoTreeView.style_data.length);
  cssInfoTreeView.tree.invalidate();

}


function initStyleSheetInfo(aElement)
{
  var stylesheet_lbox = document.getElementById('stylesheet-lbox');
  var doc_header = aElement.ownerDocument.getElementsByTagName('head')[0];
  var doc_stylesheets = aElement.ownerDocument.styleSheets; // style sheets explicitly LINKed into or
                                                            // embedded in document via STYLE element,
                                                            // but not those included via @import rules
  var lc = null, li = null;
  var fullSSList = [];
  var ieSSList = [];

  function makeStyleSheetListItem(aLabel, aMedia/*, optionalAttributes*/)
  {
    li = document.createElement('listitem');

    lc = document.createElement('listcell');
    lc.setAttribute('label', aLabel);
    li.appendChild(lc);

    lc = document.createElement('listcell');
    lc.setAttribute('label', aMedia);
    li.appendChild(lc);

    if (arguments.length > 2)
     {
       attributes = arguments[2];
       for (var i in attributes)
         li.setAttribute(i, attributes[i]);
     }

    stylesheet_lbox.appendChild(li);
  }

  // adds stylesheet and any nested stylesheets to full list
  function addStyleSheet(aStyleSheet)
  {
     // guard against duplicates and infinite loops
     for(var i=0; i < fullSSList.length; i++)
       if (aStyleSheet.href != null && aStyleSheet.href == fullSSList[i].href)
         return;
    
     var ssObj = {};
     ssObj.href = aStyleSheet.href || "<style>...</style>";
     ssObj.media = aStyleSheet.media.mediaText;
     ssObj.value = (aStyleSheet.href ? "" : aStyleSheet.ownerNode.innerHTML);

     // XXX hack to guard against empty <style> elements     
     if (aStyleSheet.href == null && aStyleSheet.ownerNode.innerHTML == "") ssObj.value = " ";

     fullSSList.push(ssObj);

     var cssRules = aStyleSheet.cssRules;
     for (var i=0; i < cssRules.length; i++)
       if (cssRules.item(i).type == CSSRule.IMPORT_RULE)
         addStyleSheet(cssRules.item(i).styleSheet);
  }


  for (var ss=0; ss < doc_stylesheets.length; ss++)
   {
     addStyleSheet(doc_stylesheets.item(ss));
   }

  for (var ss=0; ss < fullSSList.length; ss++)
   {
     makeStyleSheetListItem(fullSSList[ss].href, fullSSList[ss].media, {'value': fullSSList[ss].value});
   }

  // extract conditional ie stylesheet links, if any

/* XXX RETAINED ONLY FOR REFERENCE - TO BE REMOVED IN NEXT RELEASE
 ******************************************************************
  for (var ni=0; ni < doc_header.childNodes.length; ni++)
    if (doc_header.childNodes.item(ni).nodeType == Node.COMMENT_NODE)
     {
       var hdr_comment = doc_header.childNodes.item(ni).nodeValue;
       var m_array = hdr_comment.match(/\[if.*\]>(.*)<!\[endif\]/);
       if (m_array)
        {
          var body = aElement.ownerDocument.createElement('body');
          body.innerHTML = m_array[1];
          var ie_conditional_tag = body.firstChild.tagName.toLowerCase();
          if (ie_conditional_tag == 'link' || ie_conditional_tag == 'style')
           {
             var aLink = body.firstChild;

             var ssObj = {};
             ssObj.href = aLink.href || "<style>...</style>";
             ssObj.media = aLink.media;
             ssObj.value = (aLink.href ? "" : aLink.innerHTML);
             ieSSList.push(ssObj);
           }
        }
     }
 ******************************************************************
*/

  // see http://www.nabble.com/How-to-get-comments-in-HTML-document--td10698002.html
  // also see: https://developer.mozilla.org/en/DOM/document.evaluate
  var comments = aElement.ownerDocument.evaluate(
                 "//comment()",
                 aElement.ownerDocument,
                 null,
                 XPathResult.ORDERED_NODE_SNAPSHOT_TYPE,
                 null);

  for (var i=0; i < comments.snapshotLength; i++)
   {
     var comment_text = comments.snapshotItem(i).nodeValue;
     var m_array = comment_text.match(/\[if.*\]>([\d\D]*)<!\[endif\]/);
     if (m_array)
      {
        // rather than parse conditional comment text, we'll use FF
        // to convert it to DOM structure that can be manipulated
        var body = aElement.ownerDocument.createElement('body');
        body.innerHTML = m_array[1];

        // XXX rewrite to handle conditional comments containing multiple link or style elements
        var style_elements = body.getElementsByTagName('link');
        if (style_elements.length < 1) style_elements = body.getElementsByTagName('style');
        if (style_elements.length > 0)
         {
           var style_element = style_elements[0];

           var ssObj = {};
           ssObj.href = style_element.href || "<style>...</style>";
           ssObj.media = style_element.media;
           ssObj.value = (style_element.href ? "" : style_element.innerHTML);
           ieSSList.push(ssObj);
         }
      }
   }

   
  if (ieSSList.length > 0)
   {
     makeStyleSheetListItem('', '', {'class': "nondata"});
     makeStyleSheetListItem('ie conditional', '', {'class': "nondata"});

     for (var ss=0; ss < ieSSList.length; ss++)
       makeStyleSheetListItem(ieSSList[ss].href, ieSSList[ss].media, {'class': "ieonly", 'value': ieSSList[ss].value});
   }

  stylesheet_lbox.addEventListener('contextmenu', checkStylesheetContextMenu, true);

}


function checkStylesheetContextMenu(e)
{
  // don't display context menu for headers and blank lines

  if (document.getElementById('stylesheet-lbox').selectedItem.className == "nondata")
    e.preventDefault();
}


function openStyleSheet()
{
  // if stylesheet type is embedded, display it in a blank browser window
  // otherwise, open the stylesheet file in a new browser window

  if (document.getElementById('stylesheet-lbox').selectedItem.value != "")
   {
     var myWindowRef = window.opener.open("", "gcssamer - style element text");

/*   THIS RAISES A SECURITY EXCEPTION IN FF3.0
     myWindowRef.document.open();
     myWindowRef.write("<html><body>hello</body></html>");
     myWindowRef.close();
*/
     var blank_body = myWindowRef.document.getElementsByTagName('body')[0];
     var style_info_node = myWindowRef.document.createElement('pre');
     var style_text_node = myWindowRef.document.createTextNode(document.getElementById('stylesheet-lbox').selectedItem.value);
     style_info_node.appendChild(style_text_node);

     blank_body.appendChild(style_info_node);
   }
  else
   {
     var ss_url = document.getElementById('stylesheet-lbox').selectedItem.firstChild.getAttribute('label');
     window.opener.open(ss_url);
   }
}


function copyStyleSheetLocation()
{
  var ss_url = document.getElementById('stylesheet-lbox').selectedItem.firstChild.getAttribute('label');
  gClipboardHelper.copyString(ss_url);
}


function copyAttributeValue()
{
  var attr_value = document.getElementById('attribute-lbox').selectedItem.childNodes[1].getAttribute('label');
  gClipboardHelper.copyString(attr_value);
}


function upToParentElement()
{
  if (document.getElementById('up-to-parent-btn').getAttribute('menuonly') == 'true')
    return

  // keep style property list scroll position fixed as we move up the DOM
  var current_first_visible_row = cssInfoTreeView.tree.getFirstVisibleRow();

  gCurrentElement = gCurrentElement.parentNode;

  initAttributeInfo(gCurrentElement);
  initCSSInfo(gCurrentElement);
  cssInfoTreeView.tree.scrollToRow(current_first_visible_row);

  if (gCurrentElement.parentNode == null || gCurrentElement.parentNode.nodeType != Node.ELEMENT_NODE)
    document.getElementById('up-to-parent-btn').setAttribute('menuonly', "true");
  else
    document.getElementById('up-to-parent-btn').setAttribute('menuonly', "false");
}


function resetCurrentElement()
{
  gCurrentElement = gOriginalTargetElement;

  initAttributeInfo(gCurrentElement);
  initCSSInfo(gCurrentElement);

  if (gCurrentElement.parentNode == null || gCurrentElement.parentNode.nodeType != Node.ELEMENT_NODE)
    document.getElementById('up-to-parent-btn').setAttribute('menuonly', "true");
  else
    document.getElementById('up-to-parent-btn').setAttribute('menuonly', "false");
}


// derived from Firefox 3.0: pageInfo.js
// define a js object to implement nsITreeView

cssInfoTreeView = {
  style_data: [],

  set rowCount(c) { throw "rowCount is a readonly property"; },
  get rowCount() { return this.style_data.length },

  clear: function()
  {
    if (this.tree)
      this.tree.rowCountChanged(0, -this.style_data.length);
    this.style_data = [];
  },

  setTree: function(tree)
  {
    this.tree = tree;
  },

  getCellText: function(row, column)
  {
    return (column.id != "propselected" ? this.style_data[row][COL_PROPSTRING] + ": " +
                                          this.style_data[row][COL_PROPVALUE] : "");
  },

  setCellValue: function(row, column, value){},

  setCellText: function(row, column, value){},

  rowCountChanged: function(index, count){},

  invalidate: function()
  {
    this.tree.invalidate();
  },

  handleCopy: function(row)
  {
    return (this.style_data[row][COL_PROPSTRING] + ": " + this.style_data[row][COL_PROPVALUE] + ";");
  },

  performActionOnRow: function(action, row)
  {
    if (action == "copy") {
      var data = this.handleCopy(row)

      // per nsITreeView API, return type for this method is void, so make data available via tree attribute
      this.tree.treeBody.parentNode.setAttribute("copybuffer", data);
    }
  },

  getRowProperties: function(row, prop) { },

  getCellProperties: function(row,col,props)
  {
    if (this.style_data[row][COL_SELECTEDFLAG]){
      var aserv=Components.classes["@mozilla.org/atom-service;1"].
              getService(Components.interfaces.nsIAtomService);
      props.AppendElement(aserv.getAtom("on"));
    }
  },

  cycleCell: function(row, column)
  {
    this.style_data[row][COL_SELECTEDFLAG] = !this.style_data[row][COL_SELECTEDFLAG];

    var favorite_props = [];
    for (var idx=0; idx < this.style_data.length; idx++)
     {
       if (this.style_data[idx][COL_SELECTEDFLAG] == true)
         favorite_props.push(this.style_data[idx][COL_PROPSTRING])
     }
    gGCSSAMERUserPrefUtils.setPref('cssprops', favorite_props.join(';'));
  },

  getColumnProperties: function(column, prop) { },
  isContainer: function(index) { return false; },
  isContainerOpen: function(index) { return false; },
  isSeparator: function(index) { return false; },
  isSorted: function() { },
  canDrop: function(index, orientation) { return false; },
  drop: function(row, orientation) { return false; },
  getParentIndex: function(index) { return 0; },
  hasNextSibling: function(index, after) { return false; },
  getLevel: function(index) { return 0; },
  getImageSrc: function(row, column) { },
  getProgressMode: function(row, column) { },
  getCellValue: function(row, column) { },
  toggleOpenState: function(index) { },
  cycleHeader: function(col) { },
  selectionChanged: function() { },
  isEditable: function(row, column) { return false; },
  isSelectable: function(row, column) { return false; },
  performAction: function(action) { },
  performActionOnCell: function(action, row, column) { }
};


// from Firefox 3.0: pageInfo.js
// generic tree row copy handler

function doCopy()
{
  if (!gClipboardHelper)
    return;

  var elem = document.commandDispatcher.focusedElement;

  if (elem && "treeBoxObject" in elem) {
    var view = elem.view;
    var selection = view.selection;
    var text = [], tmp = '';
    var min = {}, max = {};

    var count = selection.getRangeCount();

    for (var i = 0; i < count; i++) {
      selection.getRangeAt(i, min, max);

      for (var row = min.value; row <= max.value; row++) {
        view.performActionOnRow("copy", row);

        tmp = elem.getAttribute("copybuffer");
        if (tmp)
          text.push(tmp);
        elem.removeAttribute("copybuffer");
      }
    }
    gClipboardHelper.copyString(text.join("\n"));
  }
}

